#ifndef _RHEO_QUADRATURE_H
#define _RHEO_QUADRATURE_H
///
/// This file is part of Rheolef.
///
/// Copyright (C) 2000-2009 Pierre Saramito <Pierre.Saramito@imag.fr>
///
/// Rheolef is free software; you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation; either version 2 of the License, or
/// (at your option) any later version.
///
/// Rheolef is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with Rheolef; if not, write to the Free Software
/// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
/// 
/// =========================================================================

#include "rheolef/reference_element.h"
#include "rheolef/point.h"

namespace rheolef { 

/*Class:quadrature
NAME: @code{quadrature} - quadrature formulae on the reference lement
@cindex  quadrature formulae
@clindex quadrature
@clindex reference_element
SYNOPSYS:
@noindent
The @code{quadrature} class defines a container for a quadrature
formulae on the reference element (@pxref{reference_element iclass}).
This container stores the nodes coordinates and the weights.
The constructor takes two arguments: the reference element 
@math{K}
and the order @math{r} of the quadrature formulae.
The formulae is exact when computing the integral
of a polynom @math{p} that degree is less or equal to order @math{r}.
@example
                  n
    /            ___
    | p(x) dx =  \    p(x_q) w_q
    / K          /__
                 q=1
@end example
LIMITATIONS:
@noindent
The formulae is optimal when it uses a minimal number
of nodes @math{n}.
Optimal quadrature formula are hard-coded in this class.
Not all reference elements and orders are yet 
implemented. This class will be completed in the future.

AUTHORS:
    LMC-IMAG, 38041 Grenoble cedex 9, France
   | Pierre.Saramito@imag.fr
DATE:   30 november 2003
End:
*/

class quadrature_option_type {
public:
// typedefs:

  typedef size_t size_type;

  typedef enum {
	gauss		= 0,
	gauss_lobatto	= 1,
	gauss_radau	= 2,
	middle_edge	= 3,
	superconvergent	= 4,
	max_family	= 5
  } family_type; // update also family_name[] in quatrature.cc

// allocators:

  quadrature_option_type(
        family_type ft = quadrature_option_type::gauss,
        size_type k = std::numeric_limits<size_type>::max());

  quadrature_option_type (const quadrature_option_type& qopt);
  quadrature_option_type& operator= (const quadrature_option_type& qopt);

// accessors & modifiers:

  size_t         get_order() const;
  family_type    get_family() const;
  std::string    get_family_name() const;
  void set_order (size_t r);
  void set_family (family_type ft);
// data:
protected:
  family_type   _family;
  size_t        _order;
};
template<class T>
struct weighted_point {
  weighted_point () : x(), w(0) {}
  weighted_point (const point_basic<T>& x1, const T& w1) : x(x1), w(w1) {}
  point_basic<T> x;
  T              w;
};
template<class T>
class quadrature_on_geo : public std::vector<weighted_point<T> > {
public:

// typedefs:

    typedef std::vector<weighted_point<T> > base;
    typedef typename base::size_type        size_type;
    typedef point_basic<T>                  x;

// alocators/deallocators:

    quadrature_on_geo ();

// modifier:

    void initialize       (reference_element hat_K, quadrature_option_type opt);
    void init_point       (quadrature_option_type opt);
    void init_edge        (quadrature_option_type opt);
    void init_triangle    (quadrature_option_type opt);
    void init_square      (quadrature_option_type opt);
    void init_tetrahedron (quadrature_option_type opt);
    void init_prism       (quadrature_option_type opt);
    void init_hexahedron  (quadrature_option_type opt);

    void wx (const point_basic<T>& x, const T& w) {
	    base::push_back (weighted_point<T>(x,w)); }

    static size_type n_node_gauss (size_type r);
    
    template<class U>
    friend std::ostream& operator<< (std::ostream&, const quadrature_on_geo<U>&);
private:
    quadrature_on_geo (const quadrature_on_geo&);
    quadrature_on_geo operator= (const quadrature_on_geo&);
};
//<quadrature:
template<class T>
class quadrature {
public:

// typedefs:

    typedef typename quadrature_on_geo<T>::size_type                 size_type;
    typedef quadrature_option_type::family_type                      family_type;
    typedef typename std::vector<weighted_point<T> >::const_iterator const_iterator;

// allocators:

    quadrature (quadrature_option_type opt = quadrature_option_type());

// modifiers:

    void set_order  (size_type order);
    void set_family (family_type ft);

// accessors:

    size_type      get_order() const;
    family_type    get_family() const;
    std::string    get_family_name() const;
    size_type      size  (reference_element hat_K) const;
    const_iterator begin (reference_element hat_K) const;
    const_iterator end   (reference_element hat_K) const;
    template<class U>
    friend std::ostream& operator<< (std::ostream&, const quadrature<U>&);
protected:
    quadrature_option_type        _options;
    mutable quadrature_on_geo<T>  _quad [reference_element::max_variant];
    mutable std::vector<bool>     _initialized;
    void _initialize (reference_element hat_K) const;
private:
    quadrature (const quadrature<T>&);
    quadrature operator= (const quadrature<T>&);
};
//>quadrature:
// ------------------------------------------------------------
// inlined
// ------------------------------------------------------------
inline 
quadrature_option_type::quadrature_option_type (family_type ft, size_type k)
 : _family(ft),
   _order(k)
{
}
inline
quadrature_option_type::quadrature_option_type (const quadrature_option_type& qopt)
 : _family(qopt._family),
   _order(qopt._order)
{
}
inline
quadrature_option_type&
quadrature_option_type::operator= (const quadrature_option_type& qopt)
{
  _family = qopt._family;
  _order  = qopt._order;
  return *this;
}
inline
quadrature_option_type::size_type
quadrature_option_type::get_order () const
{
    return _order;
}
inline
quadrature_option_type::family_type
quadrature_option_type::get_family () const
{
    return _family;
}
inline
void
quadrature_option_type::set_order (size_t r)
{
    _order = r;
}
inline
void
quadrature_option_type::set_family (family_type ft)
{
    _family = ft;
}
template<class T>
inline 
quadrature_on_geo<T>::quadrature_on_geo ()
 : std::vector<weighted_point<T> >()
{
}
template<class T>
inline
typename quadrature_on_geo<T>::size_type
quadrature_on_geo<T>::n_node_gauss (size_type r)
{
    // when using n nodes : gauss quadrature formulae order is r=2*n-1
    if (r == 0) return 1;
    size_type n = (r % 2 == 0) ? r/2+1 : (r+1)/2;
    return std::max(size_t(1), n);
}
template<class T>
inline 
typename quadrature<T>::size_type
quadrature<T>::get_order () const
{
    return _options.get_order();
}
template<class T>
inline 
typename quadrature<T>::family_type
quadrature<T>::get_family () const
{
    return _options.get_family();
}
template<class T>
inline 
std::string
quadrature<T>::get_family_name () const
{
    return _options.get_family_name();
}
template<class T>
inline 
quadrature<T>::quadrature (quadrature_option_type opt)
 : _options(opt),
   _quad(),
   _initialized(reference_element::max_variant, false)
{
}

template<class T>
inline
void
quadrature<T>::set_order (size_type r)
{
  if (get_order() != r) {
    // do not re-initialize nodes-weights if unchanged
    _options.set_order(r);
    std::fill (_initialized.begin(), _initialized.end(), false);
  }
}
template<class T>
inline
void
quadrature<T>::set_family (family_type ft)
{
  if (get_family() != ft) {
    // do not re-initialize nodes-weights if unchanged
    _options.set_family(ft);
    std::fill (_initialized.begin(), _initialized.end(), false);
  }
}

}// namespace rheolef
#endif // _RHEO_QUADRATURE_H
